/*
 * Copyright (c) 1998-2004 Matthijs Hollemans
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

#include <InterfaceKit.h>
#include <Path.h>
#include <Roster.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Grepper.h"
#include "GrepWindow.h"
#include "Model.h"

////////////////////////////////////////////////////////////////////////////////

class ResultItem : public BStringItem
{
public:

	ResultItem(entry_ref& ref) : BStringItem("", 0, false)
	{
		this->ref = ref;
		BEntry entry(&ref);
		BPath path(&entry);
		SetText(path.Path()); 
	}

	entry_ref ref;
};

////////////////////////////////////////////////////////////////////////////////

GrepWindow::GrepWindow()
	: BWindow(BRect(0, 0, 1, 1), NULL, B_DOCUMENT_WINDOW, 0)
{
	oldPattern = NULL;
	grepper = NULL;

	model.target = this;

	SetWindowTitle();
	CreateOptionsMenu();
	CreateViews();
	LayoutViews();
	LoadPrefs();
	Show(); 	
}

////////////////////////////////////////////////////////////////////////////////

GrepWindow::~GrepWindow()
{
	if (model.state == STATE_SEARCH)
	{
		grepper->Cancel();
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::FrameResized(float width, float height)
{
	super::FrameResized(width, height);
	model.frame = Frame();
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::FrameMoved(BPoint origin)
{
	super::FrameMoved(origin);
	model.frame = Frame();
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::MenusBeginning()
{
	model.FillHistoryMenu(historyMenu->Menu());
	super::MenusBeginning();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::MenusEnded()
{
	for (int32 t = historyMenu->Menu()->CountItems(); t > 0; --t)
	{
		delete historyMenu->Menu()->RemoveItem(t - 1);
	}

	super::MenusEnded();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::MessageReceived(BMessage* msg)
{
	switch (msg->what)
	{
		case MSG_START_CANCEL:     OnStartCancel();       break;
		case MSG_SEARCH_FINISHED:  OnSearchFinished();    break;
		case MSG_REPORT_FILE_NAME: OnReportFileName(msg); break;
		case MSG_REPORT_RESULT:    OnReportResult(msg);   break;
		case MSG_REPORT_ERROR:     OnReportError(msg);    break;
		case MSG_RECURSE_LINKS:    OnRecurseLinks();      break;
		case MSG_RECURSE_DIRS:     OnRecurseDirs();       break;
		case MSG_CASE_SENSITIVE:   OnCaseSensitive();     break;
		case MSG_ESCAPE_TEXT:      OnEscapeText();        break;
		case MSG_TEXT_ONLY:        OnTextOnly();          break;
		case MSG_INVOKE_PE:        OnInvokePe();          break;
		case MSG_SHOW_CONTENTS:    OnShowContents();      break;
		case MSG_INVOKE_ITEM:      OnInvokeItem();        break;
		case MSG_SEARCH_TEXT:      OnSearchText();        break;
		case MSG_SELECT_HISTORY:   OnHistoryItem(msg);    break;

		default: super::MessageReceived(msg); break;
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::SetWindowTitle()
{
	BEntry entry(&model.directory, true);
	if (entry.InitCheck() == B_OK)
	{
		BPath path;
		if (entry.GetPath(&path) == B_OK)
		{
			char title[B_PATH_NAME_LENGTH + 32];
			sprintf(title, "Tracker Grep: %s", path.Path());
			SetTitle(title);
			return;
		}
	}

	SetTitle("Tracker Grep " VERSION);
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::CreateOptionsMenu()
{
	BMenu* menu = new BMenu("Options");

	recurseLinks = new BMenuItem(
		"Follow symbolic links", new BMessage(MSG_RECURSE_LINKS));

	recurseDirs = new BMenuItem(
		"Look in sub-directories", new BMessage(MSG_RECURSE_DIRS));

	caseSensitive = new BMenuItem(
		"Case sensitive", new BMessage(MSG_CASE_SENSITIVE));

	escapeText = new BMenuItem(
		"Escape search text", new BMessage(MSG_ESCAPE_TEXT));

	textOnly = new BMenuItem(
		"Text files only", new BMessage(MSG_TEXT_ONLY));

	invokePe = new BMenuItem(
		"Open files in Pe", new BMessage(MSG_INVOKE_PE));

	menu->AddItem(recurseLinks);
	menu->AddItem(recurseDirs);
	menu->AddItem(caseSensitive);
	menu->AddItem(escapeText);
	menu->AddItem(textOnly);
	menu->AddItem(invokePe);

	optionsMenu = new BMenuField(
		BRect(0, 0, 80, 17), "optionsMenu", NULL, menu, false,
		B_FOLLOW_LEFT | B_FOLLOW_TOP);

	optionsMenu->ResizeToPreferred();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::CreateViews()
{
	// The search pattern entry field does not send a message when 
	// <Enter> is pressed, because the "Search/Cancel" button already 
	// does this and we don't want to send the same message twice.

	searchText = new BTextControl(
		BRect(0, 0, 0, 1), "SearchText", NULL, NULL, NULL,
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
		B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE); 
	
	searchText->TextView()->SetMaxBytes(1000); 
	searchText->ResizeToPreferred();  // because it doesn't have a label

	searchText->SetModificationMessage(new BMessage(MSG_SEARCH_TEXT));

	historyMenu = new BMenuField(
		BRect(0, 0, 17, 17), "historyMenu", NULL, new BMenu(""), 
		true, B_FOLLOW_RIGHT | B_FOLLOW_TOP);

	historyMenu->ResizeToPreferred();

	button = new BButton(
		BRect(0, 1, 80, 1), "Button", "Search",
		new BMessage(MSG_START_CANCEL), B_FOLLOW_RIGHT);
	
	button->MakeDefault(true);
	button->ResizeToPreferred();
	button->SetEnabled(false);
	
	showContents = new BCheckBox(
		BRect(0, 0, 1, 1), "ShowContents", "Show contents",
		new BMessage(MSG_SHOW_CONTENTS), B_FOLLOW_LEFT);
	
	showContents->SetValue(B_CONTROL_ON);
	showContents->ResizeToPreferred();
	
	searchResults = new BOutlineListView(
		BRect(0, 0, 40, 80), "SearchResults", B_SINGLE_SELECTION_LIST,
		B_FOLLOW_ALL_SIDES, B_WILL_DRAW); 

	searchResults->SetInvocationMessage(new BMessage(MSG_INVOKE_ITEM));
	searchResults->ResizeToPreferred();
}	

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::LayoutViews()
{
	BBox* background = new BBox(
		BRect(0, 0, 1, 1), B_EMPTY_STRING, 
		B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP, 
		B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE_JUMP, 
		B_PLAIN_BORDER); 

	BScrollView* scroller = new BScrollView(
		"ScrollSearchResults", searchResults, B_FOLLOW_ALL_SIDES, 
		B_FULL_UPDATE_ON_RESIZE, true, true, B_NO_BORDER);

	scroller->ResizeToPreferred();

	float width = 8 + showContents->Frame().Width()
		+ 8 + optionsMenu->Frame().Width()
		+ 8 + button->Frame().Width() + 8;

	float height = 8 + searchText->Frame().Height() + 8
		+ button->Frame().Height() + 8 + scroller->Frame().Height();

	float backgroundHeight = 8 + searchText->Frame().Height()
		+ 8 + button->Frame().Height() + 8;

	ResizeTo(width, height);  

	AddChild(background);
	background->ResizeTo(width,	backgroundHeight);
	background->AddChild(searchText);
	background->AddChild(historyMenu);
	background->AddChild(showContents);
	background->AddChild(optionsMenu);
	background->AddChild(button);

	searchText->MoveTo(8, 8);
	searchText->ResizeBy(width - historyMenu->Frame().Width() - 24, 0);
	searchText->MakeFocus(true);

	historyMenu->MoveTo(
		width - 8 - historyMenu->Frame().Width(),
		8 + (searchText->Frame().Height() - historyMenu->Frame().Height())/2);

	showContents->MoveTo(
		8, 8 + searchText->Frame().Height() + 8
			+ (button->Frame().Height() - showContents->Frame().Height())/2);

	optionsMenu->MoveTo(
		8 + showContents->Frame().Width() + 8,
		8 + searchText->Frame().Height() + 8
			+ (button->Frame().Height() - optionsMenu->Frame().Height())/2);

	button->MoveTo(
		width - button->Frame().Width() - 8,
		8 + searchText->Frame().Height() + 8);

	AddChild(scroller);
	scroller->MoveTo(0, backgroundHeight + 1);
	scroller->ResizeTo(width + 1, height - backgroundHeight);

	BRect screenRect = BScreen(this).Frame();

	MoveTo(
		(screenRect.Width() - width) / 2,
		(screenRect.Height() - height) / 2);

	SetSizeLimits(width, 10000, height, 10000);
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::LoadPrefs()
{
	Lock();

	model.LoadPrefs();

	recurseDirs->SetMarked(model.recurseDirs);
	recurseLinks->SetMarked(model.recurseLinks);
	caseSensitive->SetMarked(model.caseSensitive);
	escapeText->SetMarked(model.escapeText);
	textOnly->SetMarked(model.textOnly);
	invokePe->SetMarked(model.invokePe);

	showContents->SetValue(
		model.showContents ? B_CONTROL_ON : B_CONTROL_OFF);

	MoveTo(model.frame.left, model.frame.top);
	ResizeTo(model.frame.Width(), model.frame.Height());

	Unlock();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::SavePrefs()
{
	model.SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnStartCancel()
{
	if (model.state == STATE_IDLE)
	{
		model.state = STATE_SEARCH;

		searchResults->MakeEmpty();
		
		if (searchText->TextView()->TextLength() == 0) { return; }

		model.AddToHistory(searchText->Text());

		// From now on, we don't want to be notified when the
		// search pattern changes, because the control will be
		// displaying the names of the files we are grepping.

		searchText->SetModificationMessage(NULL);

		historyMenu->SetEnabled(false);
		optionsMenu->SetEnabled(false);
		searchText->SetEnabled(false);

		button->MakeFocus(true);
		button->SetLabel("Cancel");

		// We need to remember the search pattern, because during 
		// the grepping, the text control's text will be replaced 
		// by the name of the file that's currently being grepped. 
		// When the grepping finishes, we need to restore the old
		// search pattern.

		oldPattern = strdup(searchText->Text());

		grepper = new Grepper(searchText->Text());
		grepper->Start();
	}
	else if (model.state == STATE_SEARCH)
	{
		model.state = STATE_CANCEL;
		grepper->Cancel();
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnSearchFinished()
{
	model.state = STATE_IDLE;

	delete grepper;
	grepper = NULL;

	historyMenu->SetEnabled(true);
	optionsMenu->SetEnabled(true);

	button->SetLabel("Search");
	button->SetEnabled(true);
	
	searchText->SetEnabled(true);
	searchText->MakeFocus(true);
	searchText->SetText(oldPattern);
	searchText->TextView()->SelectAll();
	searchText->SetModificationMessage(new BMessage(MSG_SEARCH_TEXT));
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnReportFileName(BMessage* msg)
{
	searchText->SetText(msg->FindString("filename"));
}

////////////////////////////////////////////////////////////////////////////////
		
void GrepWindow::OnReportResult(BMessage* msg)
{
	entry_ref ref;
	if (msg->FindRef("ref", &ref) != B_OK) { return; }

	BListItem* item = new ResultItem(ref);
	searchResults->AddItem(item);
	item->SetExpanded(model.showContents);

	type_code type;
	int32 count;
	msg->GetInfo("text", &type, &count);

	const char* buf;
	while (msg->FindString("text", --count, &buf) == B_OK)
	{
		uchar* temp = (uchar*) strdup(buf);
		uchar* ptr  = temp;

		while (true)
		{                                // replace all
			uchar c = *ptr;              // non-printable 
			                             // characters 
			if (c == '\0') { break; }    // by spaces

			if (!(c & 0x80) && iscntrl(c)) { *ptr = ' '; }

			++ptr;
		}

		searchResults->AddUnder(
			new BStringItem((const char*) temp), item);

		free(temp);
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnReportError(BMessage* msg)
{
	const char *buf;
	if (msg->FindString("error", &buf) == B_OK)
	{
		searchResults->AddItem(new BStringItem(buf));
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnRecurseLinks()
{
	model.recurseLinks = !model.recurseLinks;
	recurseLinks->SetMarked(model.recurseLinks);
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////
		
void GrepWindow::OnRecurseDirs()
{
	model.recurseDirs = !model.recurseDirs;
	recurseDirs->SetMarked(model.recurseDirs);
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnEscapeText()
{
	model.escapeText = !model.escapeText;
	escapeText->SetMarked(model.escapeText);
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////
	
void GrepWindow::OnCaseSensitive()
{
	model.caseSensitive = !model.caseSensitive;
	caseSensitive->SetMarked(model.caseSensitive);
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////
	
void GrepWindow::OnTextOnly()
{
	model.textOnly = !model.textOnly;
	textOnly->SetMarked(model.textOnly);
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnInvokePe()
{
	model.invokePe = !model.invokePe;
	invokePe->SetMarked(model.invokePe);
	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnShowContents()
{
	model.showContents = !model.showContents;

	int32 numItems = searchResults->FullListCountItems();
	for (int32 t = 0; t < numItems; ++t)
	{
		if (model.showContents)
		{
			searchResults->Expand(searchResults->FullListItemAt(t));
		}
		else
		{
			searchResults->Collapse(searchResults->FullListItemAt(t));
		}
	}

	SavePrefs();
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OpenInPe(const entry_ref& ref, int32 lineNum)
{
	BMessage msg('Cmdl');
	msg.AddRef("refs", &ref);
	msg.AddInt32("line", lineNum);

	entry_ref pe;
	if (be_roster->FindApp("application/x-vnd.beunited.pe", &pe) == B_OK)
	{
		if (be_roster->IsRunning(&pe))
		{
			BMessenger msngr(NULL, be_roster->TeamFor(&pe));
			msngr.SendMessage(&msg);
		}
		else
		{
			be_roster->Launch(&pe, &msg);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnInvokeItem()
{
	BListItem* item = searchResults->ItemAt(
				searchResults->CurrentSelection());
	
	if (item != NULL)
	{
		int32 level = item->OutlineLevel();
		int32 lineNum = -1;

		// Get the line number.
		if (level == 1)
		{
			BStringItem* str = dynamic_cast<BStringItem*>(item);
			if (str != NULL)
			{                                  // fortunately, atol
				lineNum = atol(str->Text());   // knows when to stop
			}                                  // the conversion...
		}

		// Get the top-most item and launch its entry_ref.
		while (level != 0)
		{
			item = searchResults->Superitem(item);
			level = item->OutlineLevel();
		}

		ResultItem* entry = dynamic_cast<ResultItem*>(item);
		if (entry != NULL)
		{
			if (model.invokePe)
			{
				OpenInPe(entry->ref, lineNum);
			}
			else
			{
				be_roster->Launch(&entry->ref);
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnSearchText()
{
	button->SetEnabled(searchText->TextView()->TextLength() != 0);
}

////////////////////////////////////////////////////////////////////////////////

void GrepWindow::OnHistoryItem(BMessage* msg)
{
	const char *buf;
	if (msg->FindString("text", &buf) == B_OK)
	{
		searchText->SetText(buf);
	}
}

////////////////////////////////////////////////////////////////////////////////
