/*
 * Copyright (c) 1998-2004 Matthijs Hollemans
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

#include <Directory.h>
#include <List.h>
#include <Looper.h>
#include <NodeInfo.h>
#include <Path.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Grepper.h"
#include "Model.h"

////////////////////////////////////////////////////////////////////////////////

Grepper::Grepper(const char* pattern) 
{
	threadId = -1;
	mustQuit = false;
	currentRef = 0;

	SetPattern(pattern);

	currentDir = new BDirectory(&model.directory);
	currentDir->Rewind();

	directories = new BList(10);
	directories->AddItem(currentDir);
}

////////////////////////////////////////////////////////////////////////////////

Grepper::~Grepper()
{
	free(pattern);

	// If the thread terminated normally, then there is only 
	// one object in the list: the initial directory. But if 
	// the user aborted the search, there may be more.

	for (int32 t = directories->CountItems(); t > 0; --t)
	{
		delete directories->RemoveItem((int32) 0);
	}

	delete directories;
}

////////////////////////////////////////////////////////////////////////////////

void Grepper::Start()
{
	threadId = spawn_thread(
		SpawnThread, "GrepperThread", B_NORMAL_PRIORITY, this); 
                                
	resume_thread(threadId); 
}

////////////////////////////////////////////////////////////////////////////////

void Grepper::Cancel()
{
	mustQuit = true;
	int32 exitValue;
	wait_for_thread(threadId, &exitValue);
}

////////////////////////////////////////////////////////////////////////////////

int32 Grepper::SpawnThread(void* arg) 
{ 
	return ((Grepper*) arg)->GrepperThread();
}
   
////////////////////////////////////////////////////////////////////////////////

int32 Grepper::GrepperThread() 
{
	BMessage msg;

	char fileName[B_PATH_NAME_LENGTH]; 
	char tempString[B_PATH_NAME_LENGTH];
	char command[B_PATH_NAME_LENGTH + 32];

	BPath tempFile;
	sprintf(fileName, "/boot/var/tmp/TrackerGrep%ld", threadId);
	tempFile.SetTo(fileName);

	while (!mustQuit && GetNextName(fileName))
	{
		msg.MakeEmpty();
		msg.what = MSG_REPORT_FILE_NAME;
		msg.AddString("filename", fileName);
		model.target->PostMessage(&msg);

		msg.MakeEmpty();
		msg.what = MSG_REPORT_RESULT;
		msg.AddString("filename", fileName);
		
		BEntry entry(fileName);
		entry_ref ref;
		entry.GetRef(&ref);
		msg.AddRef("ref", &ref);

		EscapeQuotes(fileName);

		sprintf(
			command, "/boot/beos/bin/grep -hn %s %s \"%s\" > \"%s\"",
			model.caseSensitive ? "" : "-i", pattern, fileName, 
			tempFile.Path());

		int res = system(command);

		if ((res == 0) || (res == 1))
		{
			FILE* results = fopen(tempFile.Path(), "r");

			if (results != NULL)
			{
				while (fgets(tempString, B_PATH_NAME_LENGTH, results) != 0)
				{
					msg.AddString("text", tempString);
				}
		
				if (msg.HasString("text"))
				{
					model.target->PostMessage(&msg);
				}

				fclose(results);
				continue;
			}
		}

		sprintf(
			tempString, "%s: There was a problem running grep.",
			fileName);

		msg.MakeEmpty();
		msg.what = MSG_REPORT_ERROR;
		msg.AddString("error", tempString);
		model.target->PostMessage(&msg);
	}

	// We wait with removing the temporary file until after the
	// entire search has finished, to prevent a lot of flickering
	// if the Tracker window for /boot/var/tmp/ might be open.

	remove(tempFile.Path());

	msg.MakeEmpty();
	msg.what = MSG_SEARCH_FINISHED;
	model.target->PostMessage(&msg);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////

void Grepper::SetPattern(const char* src)
{
	if (model.escapeText)
	{
		// We will simply guess the size of the memory buffer
		// that we need. This should always be large enough.
		pattern = (char*) malloc((strlen(src) + 1) * 3 * sizeof(char));
		
		const char* srcPtr = src;
		char* dstPtr = pattern;

		// Put double quotes around the pattern, so separate
		// words are considered to be part of a single string.
		*dstPtr++ = '"';

		while (*srcPtr != '\0')
		{
			char c = *srcPtr++;
			
			// Put a backslash in front of characters
			// that should be escaped. 
			if ((c == '.')  || (c == ',')
			||  (c == '[')  || (c == ']')
			||  (c == '?')  || (c == '*')
			||  (c == '+')  || (c == '-')
			||  (c == ':')  || (c == '^')
			||  (c == '\'') || (c == '"'))
			{
				*dstPtr++ = '\\';
			}
			else if ((c == '\\') || (c == '$'))
			{
				// Some characters need to be escaped
				// with *three* backslashes in a row.
				*dstPtr++ = '\\';
				*dstPtr++ = '\\';
				*dstPtr++ = '\\';
			}

			// Note: we do not have to escape the
			// { } ( ) < > and | characters.

			*dstPtr++ = c;
		}

		*dstPtr++ = '"';
		*dstPtr = '\0';		
	}
	else
	{
		pattern = strdup(src);
	}
}

////////////////////////////////////////////////////////////////////////////////

void Grepper::EscapeQuotes(char* buffer)
{
	char* copy = strdup(buffer);
	for (uint32 count = 0; count < strlen(copy); ++count)
	{
		if (copy[count] == '"')
		{
			*buffer = '\\';      
			++buffer;
			*buffer = '"';
		}
		else
		{
			*buffer = copy[count];
		}
		++buffer;
	}
	*buffer = '\0';
	free(copy);
}

////////////////////////////////////////////////////////////////////////////////

bool Grepper::GetNextName(char* buffer)
{
	BEntry entry;
	struct stat fileStat;

	while (true)
	{
		// Traverse the directory to get a new BEntry.
		// GetNextEntry returns false if there are no
		// more entries, and we exit the loop.

		if (!GetNextEntry(entry)) { return false; }

		// If the entry is a subdir, then add it to the
		// list of directories and continue the loop. 
		// If the entry is a file and we can grep it
		// (i.e. it is a text file), then we're done
		// here. Otherwise, continue with the next entry.

		if (entry.GetStat(&fileStat) == B_OK) 
		{
			if (S_ISDIR(fileStat.st_mode))  // subdir
			{
				ExamineSubdir(entry);
			}
			else  // file or a (non-traversed) symbolic link
			{
				if (ExamineFile(entry, buffer)) { return true; }
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

bool Grepper::GetNextEntry(BEntry& entry)
{
	if (directories->CountItems() == 1)
	{
		return GetTopEntry(entry);
	}
	else
	{
		return GetSubEntry(entry);
	}
}

////////////////////////////////////////////////////////////////////////////////

bool Grepper::GetTopEntry(BEntry& entry)
{
	// If the user selected one or more files, we must look 
	// at the "refs" inside the message that was passed into 
	// our add-on's process_refs(). If the user didn't select 
	// any files, we will simply read all the entries from the 
	// current working directory.
	
	entry_ref fileRef;

	if (model.selectedFiles->FindRef(
			"refs", currentRef, &fileRef) == B_OK)
	{
		entry.SetTo(&fileRef, model.recurseLinks);
		++currentRef;
		return true;
	}
	else if (currentRef > 0)  // when we get here, we 
	{                         // have processed all the
		return false;         // refs from the message
	}
	else  // examine the whole directory
	{
		return currentDir->GetNextEntry(
			&entry, model.recurseLinks) == B_OK;
	}
}

////////////////////////////////////////////////////////////////////////////////

bool Grepper::GetSubEntry(BEntry& entry)
{
	if (currentDir->GetNextEntry(&entry, model.recurseLinks) != B_OK)
	{
		// If we get here, there are no more entries in 
		// this subdir, so return to the parent directory.

		directories->RemoveItem(currentDir);
		delete currentDir;
		currentDir = (BDirectory*) directories->LastItem();
	
		return GetNextEntry(entry);
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////

void Grepper::ExamineSubdir(BEntry& entry)
{
	if (model.recurseDirs)
	{
		BDirectory* dir = new BDirectory(&entry);

		if (dir->InitCheck() == B_OK)
		{
			// Add new directory to the list 
			// and make it the current one.
			directories->AddItem(dir);  
			currentDir = dir;
		}
		else
		{
			delete dir;  // clean up
		}
	}
}

////////////////////////////////////////////////////////////////////////////////

bool Grepper::ExamineFile(BEntry& entry, char* buffer)
{
	BPath path;
	if (entry.GetPath(&path) == B_OK)
	{ 
		strcpy(buffer, path.Path());

		if (!model.textOnly) { return true; }

		BNode node(&entry);
		BNodeInfo nodeInfo(&node);
		char mimeTypeString[B_MIME_TYPE_LENGTH];

		if (nodeInfo.GetType(mimeTypeString) == B_OK)
		{
			BMimeType mimeType(mimeTypeString);
			BMimeType superType;

			if (mimeType.GetSupertype(&superType) == B_OK)
			{
				if ((strcmp("text", superType.Type()) == 0) 
				||  (strcmp("message", superType.Type()) == 0))
				{							
					return true;
				} 
			}
		}
	}

	return false;
}

////////////////////////////////////////////////////////////////////////////////
