/*
 * Copyright (c) 1998-2004 Matthijs Hollemans
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a 
 * copy of this software and associated documentation files (the "Software"), 
 * to deal in the Software without restriction, including without limitation 
 * the rights to use, copy, modify, merge, publish, distribute, sublicense, 
 * and/or sell copies of the Software, and to permit persons to whom the 
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */

#include <Entry.h>
#include <File.h>
#include <List.h>
#include <MenuItem.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Model.h"

Model model;

////////////////////////////////////////////////////////////////////////////////

Model::Model()
{
	recurseDirs   = true;
	recurseLinks  = false;
	caseSensitive = false;
	escapeText    = true;
	textOnly      = true;
	invokePe      = false;
	showContents  = false;

	target = NULL;
	selectedFiles = NULL;
	state = STATE_IDLE;

	frame.left   = 100;
	frame.right  = 500;
	frame.top    = 100;
	frame.bottom = 400;
}

////////////////////////////////////////////////////////////////////////////////

void Model::LoadPrefs()
{
	BFile file(PREFS_FILE, B_READ_ONLY);

	if (file.InitCheck() != B_OK) { return; }
	if (file.Lock() != B_OK) { return; }

	int32 value;

	file.ReadAttr("RecurseDirs", B_INT32_TYPE, 0, &value, sizeof(int32));
	recurseDirs = (value != 0);

	file.ReadAttr("RecurseLinks", B_INT32_TYPE, 0, &value, sizeof(int32));
	recurseLinks = (value != 0);

	file.ReadAttr("CaseSensitive", B_INT32_TYPE, 0, &value, sizeof(int32));
	caseSensitive = (value != 0);

	file.ReadAttr("EscapeText", B_INT32_TYPE, 0, &value, sizeof(int32));
	escapeText = (value != 0);

	file.ReadAttr("TextOnly", B_INT32_TYPE, 0, &value, sizeof(int32));
	textOnly = (value != 0);

	file.ReadAttr("InvokePe", B_INT32_TYPE, 0, &value, sizeof(int32));
	invokePe = (value != 0);

	file.ReadAttr("ShowContents", B_INT32_TYPE, 0, &value, sizeof(int32));
	showContents = (value != 0);

	file.ReadAttr("WindowFrame", B_RECT_TYPE, 0, &frame, sizeof(BRect));

	file.Unlock();
}

////////////////////////////////////////////////////////////////////////////////

void Model::SavePrefs()
{
	BFile file(PREFS_FILE, B_CREATE_FILE | B_ERASE_FILE | B_WRITE_ONLY);

	if (file.InitCheck() != B_OK) { return; }
	if (file.Lock() != B_OK) { return; }

	int32 value = 2;
	file.WriteAttr("Version", B_INT32_TYPE, 0, &value, sizeof(int32));

	value = recurseDirs ? 1 : 0;
	file.WriteAttr("RecurseDirs", B_INT32_TYPE, 0, &value, sizeof(int32));

	value = recurseLinks ? 1 : 0;
	file.WriteAttr("RecurseLinks", B_INT32_TYPE, 0, &value, sizeof(int32));

	value = caseSensitive ? 1 : 0;
	file.WriteAttr("CaseSensitive", B_INT32_TYPE, 0, &value, sizeof(int32));

	value = escapeText ? 1 : 0;
	file.WriteAttr("EscapeText", B_INT32_TYPE, 0, &value, sizeof(int32));

	value = textOnly ? 1 : 0;
	file.WriteAttr("TextOnly", B_INT32_TYPE, 0, &value, sizeof(int32));
	
	value = invokePe ? 1 : 0;
	file.WriteAttr("InvokePe", B_INT32_TYPE, 0, &value, sizeof(int32));
	
	value = showContents ? 1 : 0;
	file.WriteAttr("ShowContents", B_INT32_TYPE, 0, &value, sizeof(int32));
	
	file.WriteAttr("WindowFrame", B_RECT_TYPE, 0, &frame, sizeof(BRect));

	file.Sync();
	file.Unlock();
}

////////////////////////////////////////////////////////////////////////////////

void Model::AddToHistory(const char* text) 
{
	BList* items = LoadHistory();

	for (int32 t = 0; t < items->CountItems(); ++t)
	{
		char* item = (char*) items->ItemAt(t);  // if the same text
		if (strcmp(text, item) == 0)            // is already in
		{                                       // the list, then
			free(items->RemoveItem(t));         // remove it first
			break;                              // (case-sensitive!)
		}
	}

	if (items->CountItems() == HISTORY_LIMIT)
	{
		free(items->RemoveItem((int32) 0));
	}

	items->AddItem(strdup(text));

	SaveHistory(items);
	FreeHistory(items);
}

////////////////////////////////////////////////////////////////////////////////

void Model::FillHistoryMenu(BMenu* menu)
{
	BList* items = LoadHistory();

	for (int32 t = items->CountItems(); t > 0; --t)
	{
		char* item = (char*) items->ItemAt(t - 1);
		BMessage* message = new BMessage(MSG_SELECT_HISTORY);
		message->AddString("text", item);
		menu->AddItem(new BMenuItem(item, message));
	}

	FreeHistory(items);
}

////////////////////////////////////////////////////////////////////////////////

BList* Model::LoadHistory()
{
	BList* items = new BList();

	FILE* file = fopen(HISTORY_FILE, "r");
	if (file != NULL)
	{
		while (items->CountItems() < HISTORY_LIMIT)
		{
			char line[1024 + 1];
			char* result = fgets(line, 1024, file);
			if (result == NULL) { break; }
		
			int32 pos = strlen(line) - 1;
			if (line[pos] == 10) { line[pos--] = '\0'; }  // strip off
			if (line[pos] == 13) { line[pos--] = '\0'; }  // newlines
		
			items->AddItem(strdup(result));
		}

		fclose(file);
	}

	return items;
}

////////////////////////////////////////////////////////////////////////////////

void Model::SaveHistory(BList* items)
{
	FILE* file = fopen(HISTORY_FILE, "w");
	if (file != NULL)
	{
		for (int32 t = 0; t < items->CountItems(); ++t)
		{
			fputs((char*) items->ItemAt(t), file);
			fputs("\n", file);
		}
		
		fclose(file);
	}
}

////////////////////////////////////////////////////////////////////////////////

void Model::FreeHistory(BList* items)
{
	for (int32 t = items->CountItems(); t > 0; --t)
	{
		free(items->RemoveItem(t - 1));
	}

	delete items;
}

////////////////////////////////////////////////////////////////////////////////
